"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Memberaudit Doctrine Checker
from madc import __version__


class DoctrineCheckerConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    name = "madc"
    label = "madc"
    verbose_name = f"AA Memberaudit Doctrine Checker v{__version__}"



# Standard Library
import sys
import time

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Memberaudit Doctrine Checker
from madc import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

# True when tests are currently running, else False.
IS_TESTING = sys.argv[1:2] == ["test"]


def log_timing(logs):
    """
    Ein Dekorator, der die Ausführungszeit einer Funktion misst und in die Logdatei schreibt.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            logs.debug(
                "TIME: %s run for %s seconds with args: %s",
                end_time - start_time,
                func.__name__,
                args,
            )
            return result

        return wrapper

    return decorator



# Standard Library
import json

# Django
from django.test import RequestFactory

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter
from eveuniverse.models import EveType
from memberaudit.models import Character

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.models.skillchecker import SkillList, skillvalidator
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse
from madc.tests.testdata.load_memberaudit import load_memberaudit


class TestModelsSkillChecker(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_memberaudit()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )

        cls.eve_type = EveType.objects.get(id=17940)
        cls.memberaudit = Character.objects.get(eve_character__character_id=1001)

    def test__str__(self):
        """Test should return skill checker string representation."""
        skill_checker = SkillList(
            name="Test Skill Checker",
            skill_list=json.dumps({"Mining Barge": 3}),
        )
        self.assertEqual(
            str(skill_checker),
            f"{skill_checker.name} ({skill_checker.ordering})",
        )

    def test_skillvalidator_is_valid(self):
        """Test should validate skills correctly."""
        # given
        sk_check = {
            "Mining Barge": 3,
        }

        result = skillvalidator(value=json.dumps(sk_check))

        self.assertIsNone(result)

    def test_skillvalidator_invalid_skill(self):
        """Test should raise ValidationError for invalid skill."""
        # given
        sk_check = {
            "Invalid Skill": 3,
        }

        with self.assertRaisesMessage(Exception, "Invalid Skill is not a valid skill."):
            skillvalidator(value=json.dumps(sk_check))

    def test_skillvalidator_invalid_level(self):
        """Test should raise ValidationError for invalid skill level."""
        # given
        sk_check = {
            "Mining Barge": 10,
        }

        with self.assertRaisesMessage(
            Exception, "Mining Barge level must be between 0 and 5."
        ):
            skillvalidator(value=json.dumps(sk_check))

    def test_skillvalidator_non_integer_level(self):
        """Test should raise ValidationError for non-integer skill level."""
        # given
        sk_check = {
            "Mining Barge": "high",
        }

        with self.assertRaisesMessage(
            Exception, "Mining Barge level must be an integer."
        ):
            skillvalidator(value=json.dumps(sk_check))

    def test_skillvalidator_invalid_json(self):
        """Test should raise ValidationError for invalid JSON."""
        # given
        sk_check = "Not a JSON string"

        with self.assertRaisesMessage(
            Exception, "Please provide a valid skill list in JSON format."
        ):
            skillvalidator(value=sk_check)



# Standard Library
import json

# Django
from django import template

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveType, EveTypeDogmaAttribute

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.helpers.skill_handler import SkillListHandler
from madc.models.skillchecker import SkillList

register = template.Library()

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


# pylint: disable=too-many-locals
@register.inclusion_tag("madc/partials/fittings.html", takes_context=True)
def madc_skill_overview(context) -> dict:

    # Third Party
    # pylint: disable=import-outside-toplevel
    from fittings.models import Fitting, FittingItem

    _fit = Fitting.objects.get(id=int(context.request.resolver_match.kwargs["fit_id"]))
    _items = FittingItem.objects.filter(fit=_fit).values_list("type_id", flat=True)

    _skill_ids = [182, 183, 184, 1285, 1289, 1290]
    _level_ids = [277, 278, 279, 1286, 1287, 1288]

    _types = EveTypeDogmaAttribute.objects.filter(
        eve_type_id__in=_items, eve_dogma_attribute_id__in=_skill_ids + _level_ids
    )
    _types = _types | EveTypeDogmaAttribute.objects.filter(
        eve_type_id=_fit.ship_type_type_id,
        eve_dogma_attribute_id__in=_skill_ids + _level_ids,
    )

    required = {}
    skills = {}
    sids = set()
    for t in _types:
        if t.eve_type_id not in required:
            required[t.eve_type_id] = {
                0: {"skill": 0, "level": 0},
                1: {"skill": 0, "level": 0},
                2: {"skill": 0, "level": 0},
                3: {"skill": 0, "level": 0},
                4: {"skill": 0, "level": 0},
                5: {"skill": 0, "level": 0},
            }
        a = t.eve_dogma_attribute_id
        v = t.value
        if a in _skill_ids:
            required[t.eve_type_id][_skill_ids.index(a)]["skill"] = v
        elif a in _level_ids:
            indx = _level_ids.index(a)
            if required[t.eve_type_id][indx]["level"] < v:
                required[t.eve_type_id][indx]["level"] = v

        for t in required.values():
            for skill in t.values():
                if skill["skill"]:
                    if skill["skill"] not in skills:
                        skills[skill["skill"]] = {
                            "skill": skill["skill"],
                            "level": 0,
                            "name": "",
                        }
                        sids.add(skill["skill"])
                    if skill["level"] > skills[skill["skill"]]["level"]:
                        skills[skill["skill"]]["level"] = skill["level"]
    sk_check = {}
    for t in EveType.objects.filter(id__in=list(sids)):
        skills[t.id]["name"] = t.name
        sk_check[t.name] = skills[t.id]["level"]

    char_ids = list(
        context.request.user.character_ownerships.all().values_list(
            "character__character_id", flat=True
        )[:15]
    )

    checks = SkillListHandler().check_skill_lists(
        [SkillList(name="fit", skill_list=json.dumps(sk_check))], char_ids
    )

    for __, d in checks.items():
        del d["skills"]

    response = {
        "skills": list(sorted(skills.values(), key=lambda item: item["name"])),
        "chars": dict(sorted(checks.items())),
    }
    context["skills"] = response
    return context



"""App URLs"""

# Django
from django.urls import path, re_path

# AA Memberaudit Doctrine Checker
from madc import views
from madc.api import api

app_name: str = "madc"  # pylint: disable=invalid-name

urlpatterns = [
    # -- Main URLs
    path("view/index/", views.index, name="index"),
    path(
        "<int:character_id>/view/checker/",
        views.checker,
        name="checker",
    ),
    path(
        "<int:corporation_id>/view/corporation_checker/",
        views.corporation_checker,
        name="corporation_checker",
    ),
    path(
        "<int:character_id>/view/add/",
        views.doctrine,
        name="add_doctrine",
    ),
    path(
        "<int:character_id>/view/overview/",
        views.overview,
        name="overview",
    ),
    path(
        "<int:character_id>/view/corporation_overview/",
        views.corporation_overview,
        name="corporation_overview",
    ),
    path(
        "<int:character_id>/view/administration/",
        views.administration,
        name="administration",
    ),
    # -- Administration
    path(
        "admin/add/",
        views.ajax_doctrine,
        name="ajax_doctrine",
    ),
    path(
        "admin/delete/",
        views.delete_doctrine,
        name="delete_doctrine",
    ),
    path(
        "admin/update/<int:pk>/",
        views.edit_doctrine,
        name="update_skilllist",
    ),
    # -- API System
    re_path(r"^api/", api.urls),
]



# Generated by Django 4.2.20 on 2025-07-03 10:22

# Django
import django.core.validators
from django.db import migrations, models

# AA Memberaudit Doctrine Checker
import madc.models.skillchecker


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app, Memberaudit Doctrine Checker.",
                    ),
                    ("manage_access", "Has access to moderation tools"),
                    ("admin_access", "Has access to all Administration tools"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SkillList",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_update", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(default=None, max_length=500, null=True)),
                (
                    "skill_list",
                    models.TextField(
                        default="",
                        null=True,
                        validators=[madc.models.skillchecker.skillvalidator],
                    ),
                ),
                (
                    "ordering",
                    models.IntegerField(
                        default=0,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="Order Weight",
                    ),
                ),
                (
                    "active",
                    models.BooleanField(
                        default=True,
                        help_text="If unchecked, this skill list will not be used in the skill checker.",
                        verbose_name="Active",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        default=None,
                        help_text="Category of the skill list, used for grouping.",
                        max_length=100,
                        null=True,
                        verbose_name="Category",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.20 on 2025-07-03 11:58

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("madc", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app, Memberaudit Doctrine Checker.",
                    ),
                    (
                        "corp_access",
                        "Can view Characters from own Corporation, Memberaudit Doctrine Checker.",
                    ),
                    (
                        "alliance_access",
                        "Can view Characters from own Alliance, Memberaudit Doctrine Checker.",
                    ),
                    (
                        "manage_access",
                        "Can manage this app, Memberaudit Doctrine Checker.",
                    ),
                    (
                        "admin_access",
                        "Gives full access to this app, Memberaudit Doctrine Checker.",
                    ),
                ),
            },
        ),
    ]



# Django
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Memberaudit Doctrine Checker
from madc import __title__, models

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def get_manage_permission(request, character_id):
    """Check if the user has permission to manage."""
    perms = True
    main_char = EveCharacter.objects.select_related(
        "character_ownership",
        "character_ownership__user__profile",
        "character_ownership__user__profile__main_character",
    ).get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except ObjectDoesNotExist:
        pass

    # check access
    visible = models.SkillList.objects.manage_to(request.user)
    if main_char not in visible:
        account_chars = (
            request.user.profile.main_character.character_ownership.user.character_ownerships.all()
        )
        if main_char in account_chars:
            pass
        else:
            perms = False
    return perms, main_char


def get_main_character(request, character_id) -> tuple[bool, EveCharacter | None]:
    perms = True
    main_char = EveCharacter.objects.select_related(
        "character_ownership",
        "character_ownership__user__profile",
        "character_ownership__user__profile__main_character",
    ).get(character_id=character_id)
    try:
        main_char = main_char.character_ownership.user.profile.main_character
    except ObjectDoesNotExist:
        pass

    # check access
    visible = models.SkillList.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        account_chars = (
            request.user.profile.main_character.character_ownership.user.character_ownerships.all()
        )
        if main_char in account_chars:
            pass
        else:
            perms = False
    return perms, main_char


def get_corporation(request, corporation_id) -> tuple[bool, EveCorporationInfo | None]:
    perms = True
    try:
        corporation = EveCorporationInfo.objects.get(corporation_id=corporation_id)
    except ObjectDoesNotExist:
        return False, None

    # check access
    visible = models.SkillList.objects.visible_eve_corporations(request.user)
    if corporation not in visible:
        perms = False
    return perms, corporation


def get_alts_queryset(main_char):
    try:
        linked_characters = (
            main_char.character_ownership.user.character_ownerships.all().values_list(
                "character_id", flat=True
            )
        )

        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def generate_button(pk: int, template, queryset, settings, request) -> mark_safe:
    """Generate a html button with the given template and queryset."""
    return format_html(
        render_to_string(
            template,
            {
                "pk": pk,
                "queryset": queryset,
                "settings": settings,
            },
            request=request,
        )
    )


def generate_editable_html(
    skill_list: models.SkillList,
    field_name: str,
    name: str,
    url: str,
    title: str = "",
) -> str:
    """Generate HTML for an editable field."""
    html = f"<a class='editable' href='#' data-type='text' data-pk='{skill_list.pk}' data-name='{field_name}' data-url='{url}' data-title='{str(title)}'>{name}</a>"
    return format_html(html)


def generate_editable_bool_html(
    skill_list: models.SkillList, name: str, url: str, title: str = ""
) -> str:
    """Generate HTML for a boolean field with editable functionality."""
    active_text = _("Active")
    inactive_text = _("Inactive")
    if skill_list.active:
        button_html = f"<button class='btn btn-success btn-sm'>{active_text}</button>"
    else:
        button_html = f"<button class='btn btn-danger btn-sm'>{inactive_text}</button>"

    html = f"<a class='editable-boolean no_underline' data-type='select' data-pk='{skill_list.pk}' data-name='{name}' data-url='{url}' data-title='{str(title)}' data-source='[{{\"value\": true, \"text\": \"{active_text}\"}}, {{\"value\": false, \"text\": \"{inactive_text}\"}}]' data-value='{str(skill_list.active).lower()}'>{button_html}</a>"
    return mark_safe(html)


def _collect_user_doctrines(skills_list: dict, active_skilllists: list) -> dict:
    """Collect the best doctrine skill data for the user across all their characters."""
    user_doctrines = {}

    # Process each character's skill data
    for __, character_data in skills_list.items():
        if "doctrines" not in character_data:
            continue

        for doctrine_key, doctrine_data in character_data["doctrines"].items():
            # Filter out inactive skill lists
            if doctrine_key not in active_skilllists:
                continue

            if doctrine_key not in user_doctrines:
                user_doctrines[doctrine_key] = doctrine_data
            else:
                # Take the best result (fewer missing skills) across all chars
                # Empty skills dict means all skills are trained
                current_missing_count = len(doctrine_data.get("skills", {}))
                existing_missing_count = len(
                    user_doctrines[doctrine_key].get("skills", {})
                )

                # Prefer the character with fewer missing skills (0 is best)
                if current_missing_count < existing_missing_count:
                    user_doctrines[doctrine_key] = doctrine_data

    return user_doctrines



"""TestView class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# AA Memberaudit Doctrine Checker
from madc import views
from madc.models.skillchecker import SkillList
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse


class TestAdministrationView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )
        cls.skill_plan = SkillList.objects.create(
            name="Test Skillplan",
            category="Test Category",
            active=True,
            ordering=1,
            skill_list='{"Test Level 1": 1, "Test Level 5": 5, "Test Level 3": 3}',
        )

    def test_edit_doctrine_name(self):
        """Test should edit doctrine name."""
        # given
        form_data = {"name": "name", "value": "New Doctrine Name"}
        # when
        request = self.factory.post(
            reverse("madc:update_skilllist", args=[self.skill_plan.pk]),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.edit_doctrine(request, pk=self.skill_plan.pk)
        response_data = json.loads(response.content)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response_data["message"],
            "Name updated successfully",
        )

    def test_edit_doctrine_active(self):
        """Test should edit doctrine active status."""
        # given
        form_data = {"name": "active", "value": "false"}
        # when
        request = self.factory.post(
            reverse("madc:update_skilllist", args=[self.skill_plan.pk]),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.edit_doctrine(request, pk=self.skill_plan.pk)
        response_data = json.loads(response.content)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response_data["message"],
            "Active status updated successfully",
        )
        self.skill_plan.refresh_from_db()
        self.assertFalse(self.skill_plan.active)

    def test_edit_doctrine_ordering(self):
        """Test should edit doctrine ordering."""
        # given
        form_data = {"name": "ordering", "value": "5"}
        # when
        request = self.factory.post(
            reverse("madc:update_skilllist", args=[self.skill_plan.pk]),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.edit_doctrine(request, pk=self.skill_plan.pk)
        response_data = json.loads(response.content)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response_data["message"],
            "Ordering updated successfully",
        )
        self.skill_plan.refresh_from_db()
        self.assertEqual(self.skill_plan.ordering, 5)

    def test_edit_doctrine_category(self):
        """Test should edit doctrine category."""
        # given
        form_data = {"name": "category", "value": "New Category"}
        # when
        request = self.factory.post(
            reverse("madc:update_skilllist", args=[self.skill_plan.pk]),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.edit_doctrine(request, pk=self.skill_plan.pk)
        response_data = json.loads(response.content)

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(
            response_data["message"],
            "Category updated successfully",
        )
        self.skill_plan.refresh_from_db()
        self.assertEqual(self.skill_plan.category, "New Category")

    def test_edit_doctrine_no_permission(self):
        """Test should not edit doctrine without permission."""
        # given
        form_data = {"name": "name", "value": "New Doctrine Name"}
        # when
        request = self.factory.post(
            reverse("madc:update_skilllist", args=[self.skill_plan.pk]),
            data=form_data,
        )

        request.user = self.user_no_access

        response = views.edit_doctrine(request, pk=self.skill_plan.pk)

        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)



"""TestView class."""

# Standard Library
from unittest.mock import MagicMock

# Django
from django.template import RequestContext
from django.test import RequestFactory, TestCase, modify_settings
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter
from eveuniverse.models import EveType

# AA Memberaudit Doctrine Checker
from madc.helpers.skill_handler import fittings_installed
from madc.templatetags.doctrinechecker import madc_skill_overview
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse


class TestSkillOverviewView(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )
        cls.eve_type = EveType.objects.get(id=17478)

    @modify_settings(INSTALLED_APPS={"remove": "fittings"})
    def test_fittings_installed_should_return_false(self):
        """Test should return false if fittings is not installed."""
        self.assertFalse(fittings_installed())

    @modify_settings(INSTALLED_APPS={"append": "fittings"})
    def test_fittings_installed_should_return_true(self):
        """Test should return true if fittings is installed."""
        self.assertTrue(fittings_installed())

    def test_overview_view(self):
        """Test should load overview templatetags."""
        # given
        # Third Party
        from fittings.models import Fitting

        fitting = Fitting.objects.create(
            name="Test Fitting",
            description="A fitting for testing.",
            ship_type=self.eve_type,
            ship_type_type_id=self.eve_type.id,
        )

        request = self.factory.get(reverse("madc:index"))
        # Patch resolver_match.kwargs
        request.resolver_match = MagicMock()
        request.resolver_match.kwargs = {"fit_id": fitting.pk}
        request.user = self.user
        context = RequestContext(request)
        # when
        response = madc_skill_overview(context=context)
        # then
        skill_names = []
        for ctx in response:
            if isinstance(ctx, dict) and "skills" in ctx:
                skills = ctx["skills"].get("skills", [])
                skill_names = [s.get("name") for s in skills]
        self.assertIn("Mining Barge", skill_names)



"""TestAPI class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCorporationInfo

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter
from eveuniverse.models import EveType
from memberaudit.models import Character

# AA Memberaudit Doctrine Checker
from madc.models.skillchecker import SkillList
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse
from madc.tests.testdata.load_memberaudit import load_memberaudit

API_PATH = "madc.api.admin"
API_URL = "madc:api"


class TestAdminAPI(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_memberaudit()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )

        cls.eve_type = EveType.objects.get(id=17940)
        cls.memberaudit = Character.objects.get(eve_character__character_id=1001)
        cls.skill_list = SkillList.objects.create(
            name="Test Skill List",
            category="Test Category",
            active=True,
            skill_list=json.dumps({str(cls.eve_type.name): 5}),
            ordering=1,
        )

    def test_admin_endpoint_access(self):
        # given
        url = reverse(f"{API_URL}:admin_doctrines")
        self.client.force_login(self.user_admin_access)
        # when
        response = self.client.get(url)
        response_json = response.json()
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIn("Test Skill List", response_json)
        self.assertIn(
            "Test Category", response_json["Test Skill List"]["category"]["html"]
        )

    def test_admin_endpoint_no_access(self):
        # given
        url = reverse(f"{API_URL}:admin_doctrines")
        self.client.force_login(self.user_no_access)
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_character_overview_endpoint(self):
        # given
        url = reverse(f"{API_URL}:get_character_overview")
        self.client.force_login(self.user)
        # when
        response = self.client.get(url)
        response_json = response.json()
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(len(response_json), 1)

    def test_character_overview_endpoint_admin_access(self):
        # given
        url = reverse(f"{API_URL}:get_character_overview")
        self.client.force_login(self.user_admin_access)
        # when
        response = self.client.get(url)
        response_json = response.json()

        all_characters = Character.objects.count()
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(len(response_json), all_characters)

    def test_character_overview_no_access(self):
        # given
        url = reverse(f"{API_URL}:get_character_overview")
        self.client.force_login(self.user_no_access)
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_corporation_overview_endpoint(self):
        # given
        url = reverse(f"{API_URL}:get_corporation_overview")
        self.client.force_login(self.user)
        # when
        response = self.client.get(url)
        response_json = response.json()

        counted_corps = len(response_json)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(counted_corps, 1)

    def test_corporation_overview_endpoint_admin_access(self):
        # given
        url = reverse(f"{API_URL}:get_corporation_overview")
        self.client.force_login(self.user_admin_access)
        # when
        response = self.client.get(url)
        response_json = response.json()

        counted_corps = len(response_json)
        all_corps = EveCorporationInfo.objects.count()

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(counted_corps, all_corps)

    def test_corporation_overview_no_access(self):
        # given
        url = reverse(f"{API_URL}:get_corporation_overview")
        self.client.force_login(self.user_no_access)
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)



# Standard Library
from typing import Any

# Third Party
from ninja import NinjaAPI

# Django
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from memberaudit.models import Character

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.api import schema
from madc.api.helpers import (
    generate_button,
    generate_editable_bool_html,
    generate_editable_html,
    get_manage_permission,
)
from madc.models import SkillList

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class DoctrineCheckerAdminApiEndpoints:
    tags = ["Doctrine Administration"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "administration/",
            response={200: Any, 403: str},
            tags=self.tags,
        )
        def admin_doctrines(request):
            character_id = request.user.profile.main_character.character_id
            response, __ = get_manage_permission(request, character_id)

            if not response:
                return 403, str(_("Permission Denied"))

            skilllist_obj = SkillList.objects.all().order_by("ordering", "name")

            skilllist_dict = {}

            btn_template = "madc/partials/form/button.html"
            settings_dict = {
                "title": _("Delete Skill Plan"),
                "color": "danger",
                "icon": "fa fa-trash",
                "text": _("Are you sure you want to delete this skill plan?"),
                "modal": "skillplan-delete",
                "action": reverse(
                    viewname="madc:delete_doctrine",
                ),
                "ajax": "action",
            }

            for skill_list in skilllist_obj:
                url_doctrine = reverse(
                    viewname="madc:api:get_doctrine_skills",
                    kwargs={"pk": skill_list.pk},
                )

                skills_html = f'<button class="btn btn-primary btn-sm" data-bs-toggle="modal" data-bs-target="#modalViewDoctrineContainer" data-ajax_doctrine="{url_doctrine}">{len(skill_list.get_skills())} Skills</button>'

                delete_btn = generate_button(
                    pk=skill_list.pk,
                    template=btn_template,
                    queryset=skilllist_obj,
                    settings=settings_dict,
                    request=request,
                )

                skilllist_dict[skill_list.name] = {
                    "name": {
                        "html": generate_editable_html(
                            skill_list,
                            field_name="name",
                            name=skill_list.name,
                            url=reverse(
                                viewname="madc:update_skilllist",
                                kwargs={"pk": skill_list.pk},
                            ),
                            title=_("Enter name"),
                        ),
                        "sort": skill_list.name,
                    },
                    "skills": format_html(skills_html),
                    "active": {
                        "html": generate_editable_bool_html(
                            skill_list,
                            name="active",
                            title=_("Toggle active status"),
                            url=reverse(
                                viewname="madc:update_skilllist",
                                kwargs={"pk": skill_list.pk},
                            ),
                        ),
                        "sort": skill_list.active,
                    },
                    "ordering": {
                        "html": generate_editable_html(
                            skill_list,
                            field_name="ordering",
                            name=str(skill_list.ordering),
                            url=reverse(
                                viewname="madc:update_skilllist",
                                kwargs={"pk": skill_list.pk},
                            ),
                            title=_("Enter ordering"),
                        ),
                        "sort": skill_list.ordering,
                    },
                    "category": {
                        "html": generate_editable_html(
                            skill_list,
                            field_name="category",
                            name=str(skill_list.category),
                            url=reverse(
                                viewname="madc:update_skilllist",
                                kwargs={"pk": skill_list.pk},
                            ),
                            title=_("Enter category"),
                        ),
                        "sort": skill_list.category,
                    },
                    "actions": {
                        "delete": format_html(delete_btn),
                    },
                }

            return skilllist_dict

        @api.get(
            "character/overview/",
            response={200: list[schema.CharacterOverview], 403: str},
            tags=self.tags,
        )
        def get_character_overview(request):
            chars_visible = SkillList.objects.visible_eve_characters(request.user)

            if chars_visible is None:
                return 403, "Permission Denied"

            chars_ids = chars_visible.values_list("character_id", flat=True)

            users_char_ids = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            )

            output = []

            for character in users_char_ids:
                # pylint: disable=broad-exception-caught
                try:
                    character_data = {
                        "character_id": character.main_character.character_id,
                        "character_name": character.main_character.character_name,
                        "corporation_id": character.main_character.corporation_id,
                        "corporation_name": character.main_character.corporation_name,
                        "alliance_id": character.main_character.alliance_id,
                        "alliance_name": character.main_character.alliance_name,
                    }
                    output.append({"character": character_data})
                except AttributeError:
                    continue

            return output

        @api.get(
            "corporation/overview/",
            response={200: list[schema.CorporationOverview], 403: str},
            tags=self.tags,
        )
        def get_corporation_overview(request):
            corps_visible = SkillList.objects.visible_eve_corporations(request.user)

            if corps_visible is None:
                return 403, "Permission Denied"

            # Collect unique corporation IDs from visible corporations
            corp_ids = corps_visible.distinct().values_list("corporation_id", flat=True)

            # Get existing corporation IDs from characters
            existing_corp_ids = (
                Character.objects.filter(
                    eve_character__corporation_id__in=corp_ids,
                )
                .select_related(
                    "eve_character",
                )
                .distinct()
                .values_list("eve_character__corporation_id", flat=True)
            )

            # Get Corporations from Existing Members
            corps = EveCorporationInfo.objects.filter(
                corporation_id__in=existing_corp_ids,
            )

            output = []

            for corporation in corps:
                try:
                    corporation_data = {
                        "corporation_id": corporation.corporation_id,
                        "corporation_name": corporation.corporation_name,
                        "alliance_id": getattr(
                            corporation.alliance, "alliance_id", None
                        ),
                        "alliance_name": getattr(
                            corporation.alliance, "alliance_name", "N/A"
                        ),
                    }
                    output.append({"corporation": corporation_data})
                except AttributeError:
                    continue

            return output



# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from app_utils.testing import NoSocketsTestCase

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.decorators import (
    log_timing,
)


class TestDecorators(NoSocketsTestCase):
    def test_log_timing(self):
        # given
        logger = LoggerAddTag(get_extension_logger(__name__), __title__)

        @log_timing(logger)
        def trigger_log_timing():
            return "Log Timing"

        # when
        result = trigger_log_timing()
        # then
        self.assertEqual(result, "Log Timing")



"""TestView class."""

# Standard Library
from http import HTTPStatus
from unittest.mock import patch

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# AA Memberaudit Doctrine Checker
from madc import views
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse

INDEX_PATH = "madc.views"


class TestViewIndexAccess(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )

    def test_index_view(self):
        """Test should load index view."""
        # given
        request = self.factory.get(reverse("madc:index"))
        request.user = self.user
        # when
        response = views.index(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)

    def test_checker_view(self):
        """Test should load checker view."""
        # given
        request = self.factory.get(
            reverse(
                "madc:checker", args=[self.character_ownership.character.character_id]
            )
        )
        request.user = self.user
        # when
        response = views.checker(
            request, character_id=self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_corporation_checker_view(self):
        """Test should load corporation checker view."""
        # given
        request = self.factory.get(
            reverse(
                "madc:corporation_checker",
                args=[
                    self.character_ownership_corp.character.corporation.corporation_id
                ],
            )
        )
        request.user = self.user_corp_access
        # when
        response = views.corporation_checker(
            request,
            corporation_id=self.character_ownership_corp.character.corporation.corporation_id,
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_overview_view(self):
        """Test should load overview view."""
        # given
        request = self.factory.get(
            reverse(
                "madc:overview", args=[self.character_ownership.character.character_id]
            )
        )
        request.user = self.user
        # when
        response = views.overview(
            request, character_id=self.character_ownership.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_administration_view(self):
        """Test should load administration view."""
        # given
        request = self.factory.get(
            reverse(
                "madc:administration",
                args=[self.character_ownership_admin.character.character_id],
            )
        )
        request.user = self.user_admin_access
        # when
        response = views.administration(
            request, character_id=self.character_ownership_admin.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_add_doctrine_view(self):
        """Test should load add doctrine view."""
        # given
        request = self.factory.get(
            reverse(
                "madc:add_doctrine",
                args=[self.character_ownership_admin.character.character_id],
            )
        )
        request.user = self.user_admin_access
        # when
        response = views.doctrine(
            request, character_id=self.character_ownership_admin.character.character_id
        )
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    @patch(INDEX_PATH + ".messages")
    def test_ajax_doctrine_view(self, mock_messages):
        """Test should load ajax doctrine view."""
        # given
        form_data = {
            "name": "Test Skillplan",
            "ordering": 1,
            "category": "Test Category",
            "skill_list": "Test Level 1\nTest Level 5\nTest Level 3",
        }
        # when
        request = self.factory.post(
            reverse("madc:ajax_doctrine"),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.ajax_doctrine(request)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.success.assert_called_once()

    @patch(INDEX_PATH + ".messages")
    def test_ajax_doctrine_view_error(self, mock_messages):
        """Test should load ajax doctrine view with error."""
        # given
        form_data = {}
        # when
        request = self.factory.post(
            reverse("madc:ajax_doctrine"),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.ajax_doctrine(request)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_once()

    @patch(INDEX_PATH + ".messages")
    def test_ajax_doctrine_view_wrong_parsed_skills(self, mock_messages):
        """Test should load ajax doctrine view with wrong parsed skills."""
        # given
        form_data = {
            "name": "Test Skillplan",
            "ordering": 1,
            "category": "Test Category",
            "skill_list": "Invalid Skill Format",
        }
        # when
        request = self.factory.post(
            reverse("madc:ajax_doctrine"),
            data=form_data,
        )

        request.user = self.user_admin_access

        response = views.ajax_doctrine(request)

        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        mock_messages.error.assert_called_with(
            request,
            "There was an error with your skill plan: Skill List (Skillplan format): Line 1: 'Invalid Skill Format' - Invalid format. Use 'Skill Name Level' (Level 1-5) or EVE localized format",
        )



# Generated by Django 4.2.11 on 2025-07-04 10:29

# Django
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("madc", "0002_alter_general_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="skilllist",
            options={"default_permissions": ()},
        ),
    ]



"""TestAPI class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter
from eveuniverse.models import EveType
from memberaudit.models import Character

# AA Memberaudit Doctrine Checker
from madc.api import helpers
from madc.models.skillchecker import SkillList
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse
from madc.tests.testdata.load_memberaudit import load_memberaudit

API_PATH = "madc.api.admin"
API_URL = "madc:api"


class TestAPIHelper(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_memberaudit()

        cls.factory = RequestFactory()
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1000,
                permissions=[],
            )
        )
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )

        cls.eve_type = EveType.objects.get(id=17940)
        cls.memberaudit = Character.objects.get(eve_character__character_id=1001)
        cls.skill_list = SkillList.objects.create(
            name="Test Skill Ready",
            category="Test Category",
            active=True,
            skill_list=json.dumps({str(cls.eve_type.name): 5}),
            ordering=1,
        )
        cls.skill_list2 = SkillList.objects.create(
            name="Test Skill Baby",
            active=True,
            skill_list=json.dumps({str(cls.eve_type.name): 5}),
            ordering=1,
        )

    def test_get_main_character(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        # when
        perms, main_char = helpers.get_main_character(
            request, self.character_ownership_admin.character.character_id
        )

        # then
        self.assertTrue(perms)
        self.assertEqual(
            main_char.character_id,
            self.character_ownership_admin.character.character_id,
        )

    def test_get_main_character_no_access(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_no_access

        # when
        perms, main_char = helpers.get_main_character(
            request, self.character_ownership_admin.character.character_id
        )

        # then
        self.assertFalse(perms)
        self.assertEqual(
            main_char.character_id,
            self.character_ownership_admin.character.character_id,
        )

    def test_get_corporation(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access
        corporation_id = self.character_ownership.character.corporation_id

        # when
        perms, corporation = helpers.get_corporation(request, corporation_id)

        # then
        self.assertTrue(perms)
        self.assertEqual(corporation.corporation_id, corporation_id)

    def test_get_corporation_no_access(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_no_access
        corporation_id = self.character_ownership.character.corporation_id

        # when
        perms, corporation = helpers.get_corporation(request, corporation_id)

        # then
        self.assertFalse(perms)
        self.assertEqual(corporation.corporation_id, corporation_id)

    def test_get_corporation_not_exist(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access
        corporation_id = 99999999  # Non-existent corporation ID

        # when
        perms, corporation = helpers.get_corporation(request, corporation_id)

        # then
        self.assertFalse(perms)
        self.assertIsNone(corporation)

    def test_get_alts_queryset(self):
        # when
        alts_qs = helpers.get_alts_queryset(self.character_ownership.character)

        # then
        self.assertIn(self.character_ownership.character, alts_qs)
        self.assertEqual(alts_qs.count(), 1)

    def test__collect_user_doctrines_single_char(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        skills_list = {
            "Gneuten": {
                "character_id": 1001,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1024000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1792000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "Heavy Fighters": {
                        "skill_group": "Fighters",
                        "sp_total": 819200,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                },
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
        }

        active_skilllists = [str(self.skill_list.name), str(self.skill_list2.name)]

        # when
        doctrines = helpers._collect_user_doctrines(
            skills_list=skills_list, active_skilllists=active_skilllists
        )
        # then
        self.assertIn("Test Skill Ready", doctrines)

    def test__collect_user_doctrines_multiple_chars(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        skills_list = {
            "Gneuten": {
                "character_id": 1001,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1024000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1792000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "Heavy Fighters": {
                        "skill_group": "Fighters",
                        "sp_total": 819200,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                },
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
            "rotze Rotineque": {
                "character_id": 1002,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1792000,
                        "active_level": 2,
                        "trained_level": 3,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1092000,
                        "active_level": 3,
                        "trained_level": 4,
                    },
                },
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {"XL Torpedoes": 5, "Heavy Fighters": 5},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
            "Hector Fieramosca": {
                "character_id": 1003,
                "skills": {},
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {"XL Torpedoes": 5, "Heavy Fighters": 5},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
        }

        active_skilllists = [str(self.skill_list.name), str(self.skill_list2.name)]

        # when
        doctrines = helpers._collect_user_doctrines(
            skills_list=skills_list, active_skilllists=active_skilllists
        )
        # then
        self.assertIn("Test Skill Ready", doctrines)
        self.assertNotIn("XL Torpedoes", doctrines["Test Skill Ready"]["skills"])

    def test__collect_user_doctrines_multiple_chars_missing_skills(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        skills_list = {
            "rotze Rotineque": {
                "character_id": 1002,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1792000,
                        "active_level": 2,
                        "trained_level": 3,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1092000,
                        "active_level": 3,
                        "trained_level": 4,
                    },
                },
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {"XL Torpedoes": 5, "Heavy Fighters": 5},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
            "Hector Fieramosca": {
                "character_id": 1003,
                "skills": {},
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {"XL Torpedoes": 5, "Heavy Fighters": 5},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
        }

        active_skilllists = [str(self.skill_list.name), str(self.skill_list2.name)]

        # when
        doctrines = helpers._collect_user_doctrines(
            skills_list=skills_list, active_skilllists=active_skilllists
        )
        # then
        self.assertIn("Test Skill Ready", doctrines)
        self.assertIn("XL Torpedoes", doctrines["Test Skill Ready"]["skills"])

    def test__collect_user_doctrines_no_active_skilllists(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        skills_list = {
            "Gneuten": {
                "character_id": 1001,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1024000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1792000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "Heavy Fighters": {
                        "skill_group": "Fighters",
                        "sp_total": 819200,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                },
                "doctrines": {
                    "Test Skill Ready": {
                        "skills": {},
                        "order": 0,
                        "category": "SIGMA",
                        "html": "<div class='doctrine-item btn-group' role='group' data-doctrine='Test Skill Ready'><button type='button' class='btn btn-success btn-sm'>Test Skill Ready</button><button type='button' class='flex-one btn btn-success btn-sm'><i class='fa-solid fa-check'></i></button></div>",
                    },
                },
            },
        }

        active_skilllists = []

        # when
        doctrines = helpers._collect_user_doctrines(
            skills_list=skills_list, active_skilllists=active_skilllists
        )
        # then
        self.assertEqual(doctrines, {})

    def test__collect_user_doctrines_no_doctrines(self):
        # given
        request = self.factory.get("/")
        request.user = self.user_admin_access

        skills_list = {
            "Gneuten": {
                "character_id": 1001,
                "skills": {
                    "Acceleration Control": {
                        "skill_group": "Navigation",
                        "sp_total": 1024000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "XL Torpedoes": {
                        "skill_group": "Missiles",
                        "sp_total": 1792000,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                    "Heavy Fighters": {
                        "skill_group": "Fighters",
                        "sp_total": 819200,
                        "active_level": 5,
                        "trained_level": 5,
                    },
                },
            },
        }

        active_skilllists = [str(self.skill_list.name), str(self.skill_list2.name)]

        # when
        doctrines = helpers._collect_user_doctrines(
            skills_list=skills_list, active_skilllists=active_skilllists
        )
        # then
        self.assertEqual(doctrines, {})



# Django
from django.db import models

# Alliance Auth
from allianceauth.authentication.models import User
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Memberaudit Doctrine Checker
from madc import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class SkillListQuerySet(models.QuerySet):
    def manage_to(self, user: User):
        qs = EveCharacter.objects.get_queryset()
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all Data for superuser %s.", user)
            return qs.all()

        if user.has_perm("madc.admin_access"):
            logger.debug("Returning all Data for %s.", user)
            return qs.all()

        if user.has_perm("madc.manage_access"):
            try:
                char = user.profile.main_character
                char: EveCharacter

                assert char
                # build all accepted queries
                queries = [models.Q(character_ownership__user=user)]

                if user.has_perm("madc.corp_access"):
                    queries.append(models.Q(corporation_id=char.corporation_id))

                if user.has_perm("madc.alliance_access"):
                    queries.append(models.Q(alliance_id=char.alliance_id))

                logger.debug(
                    "%s queries for user %s visible chracters.", len(queries), user
                )
                # filter based on queries
                query = queries.pop()
                for q in queries:
                    query |= q
                return qs.filter(query)
            except AssertionError:
                logger.debug("User %s has no main character. Nothing visible.", user)
                return qs.none()

        logger.debug("User %s has no permission. Nothing visible.", user)
        return qs.none()


class SkillListManagerBase(models.Manager):
    def get_queryset(self):
        return SkillListQuerySet(self.model, using=self._db)

    @staticmethod
    def visible_eve_characters(user: User):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("madc.admin_access"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            char: EveCharacter

            assert char
            # build all accepted queries
            queries = [models.Q(character_ownership__user=user)]

            if user.has_perm("madc.corp_access"):
                queries.append(models.Q(corporation_id=char.corporation_id))

            if user.has_perm("madc.alliance_access"):
                queries.append(models.Q(alliance_id=char.alliance_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )
            # filter based on queries
            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    @staticmethod
    def visible_eve_corporations(user: User):
        qs = EveCorporationInfo.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all corporations for superuser %s.", user)
            return qs.all()

        if user.has_perm("madc.admin_access"):
            logger.debug("Returning all corporations for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            char: EveCharacter

            assert char
            # build all accepted queries
            queries = [models.Q(corporation_id=char.corporation_id)]

            if user.has_perm("madc.corp_access"):
                queries.append(models.Q(corporation_id=char.corporation_id))

            if user.has_perm("madc.alliance_access"):
                queries.append(models.Q(alliance_id=char.alliance_id))

            logger.debug(
                "%s queries for user %s visible corporations.", len(queries), user
            )
            # filter based on queries
            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)


SkillListManager = SkillListManagerBase.from_queryset(SkillListQuerySet)



"""Shared ESI client for Memberaudit Doctrine Checker."""

# Alliance Auth
from esi.clients import EsiClientProvider

# AA Memberaudit Doctrine Checker
from madc import __app_name_useragent__, __github_url__, __version__
from madc.helpers.skill_handler import SkillListHandler

esi = EsiClientProvider(
    ua_appname=__app_name_useragent__, ua_version=__version__, ua_url=__github_url__
)
skills = SkillListHandler()



# Standard Library
import json

# Django
from django.test import RequestFactory

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter
from eveuniverse.models import EveType
from memberaudit.models import Character, CharacterSkill

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.helpers.skill_handler import SkillListHandler
from madc.models import SkillList
from madc.tests.testdata.load_allianceauth import load_allianceauth
from madc.tests.testdata.load_eveuniverse import load_eveuniverse
from madc.tests.testdata.load_memberaudit import load_memberaudit


class TestSkillHandler(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_memberaudit()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=["madc.basic_access"],
        )
        cls.user_corp_access, cls.character_ownership_corp = (
            create_user_from_evecharacter(
                character_id=1002,
                permissions=["madc.basic_access", "madc.corp_access"],
            )
        )
        cls.user_admin_access, cls.character_ownership_admin = (
            create_user_from_evecharacter(
                character_id=1003,
                permissions=[
                    "madc.basic_access",
                    "madc.corp_access",
                    "madc.admin_access",
                ],
            )
        )
        cls.user_no_access, cls.character_ownership_no_access = (
            create_user_from_evecharacter(
                character_id=1004,
                permissions=[],
            )
        )

        cls.eve_type = EveType.objects.get(id=17940)
        cls.memberaudit = Character.objects.get(eve_character__character_id=1001)

    def test_get_skill_list_hash(self):
        """Test should generate correct skill list hash."""
        # given
        handler = SkillListHandler()
        skills = ["Skill A", "Skill B", "Skill C"]
        expected_hash = "3d6a81fde9ea9a3e2e791536378c37f6"
        # then
        self.assertEqual(handler._get_skill_list_hash(skills), expected_hash)

    def test_get_chars_hash(self):
        """Test should generate correct character hash."""
        # given
        handler = SkillListHandler()
        characters = [1001, 1002, 1003]
        expected_hash = "fee1f06f37321e04f41a8caaa2f23330"
        # then
        self.assertEqual(handler._get_chars_hash(characters), expected_hash)

    def test_generate_doctrine_html(self):
        """Test should generate doctrine HTML snippet."""
        # given
        handler = SkillListHandler()
        # when
        html = handler._generate_doctrine_html(
            skill_list_name="Test Doctrine",
            character_id=1001,
            has_missing_skills=True,
            skill_list_pk=1,
        )
        # then
        self.assertIn("Test Doctrine", html)
        self.assertIn("1001", html)
        self.assertIn("btn-danger", html)
        self.assertIn("<i class='fa-solid fa-copy'></i>", html)

    def test_generate_doctrine_html_no_missing(self):
        """Test should generate doctrine HTML snippet for complete skills."""
        # given
        handler = SkillListHandler()
        # when
        html = handler._generate_doctrine_html(
            skill_list_name="Test Doctrine",
            character_id=1001,
            has_missing_skills=False,
        )
        # then
        self.assertIn("Test Doctrine", html)
        self.assertIn("btn-success", html)
        self.assertIn("<i class='fa-solid fa-check'></i>", html)

    def test_check_skill_lists(self):
        """Test should check skill lists for linked characters."""
        # given
        sk_check = {
            "Skill II": 2,
            "Skill I": 1,
        }

        skill_list = SkillList(name="fit", skill_list=json.dumps(sk_check))

        CharacterSkill.objects.create(
            character=self.memberaudit,
            eve_type=self.eve_type,
            active_skill_level=3,
            skillpoints_in_skill=900000,
            trained_skill_level=5,
        )

        handler = SkillListHandler()
        linked_characters = [self.character_ownership.character.character_id]
        # when
        result = handler.check_skill_lists(
            skill_lists=[skill_list],
            linked_characters=linked_characters,
        )

        character_name = self.memberaudit.eve_character.character_name
        # then
        self.assertIn(character_name, result)
        self.assertIn("fit", result[character_name]["doctrines"])
        self.assertTrue(result[character_name]["doctrines"]["fit"]["html"])
        self.assertIn("Mining Barge", result[character_name]["skills"])

    def test_check_skill_lists_no_skills(self):
        """Test should check skill lists for linked characters with no skills."""
        # given
        sk_check = {
            "Skill II": 2,
            "Skill I": 1,
        }

        skill_list = SkillList(name="fit", skill_list=json.dumps(sk_check))

        handler = SkillListHandler()
        linked_characters = [self.character_ownership.character.character_id]
        # when
        result = handler.check_skill_lists(
            skill_lists=[skill_list],
            linked_characters=linked_characters,
        )
        # then
        self.assertEqual({}, result)

    def test_get_users_skill_lists(self):
        """Test should get users skill lists."""
        # given
        handler = SkillListHandler()
        # when
        result = handler.get_users_skill_list([self.user])
        # then
        self.assertIn(self.user.pk, result)
        self.assertIn(
            self.memberaudit.eve_character.character_id, result[self.user.pk]["chars"]
        )
        self.assertIn("doctrines", result[self.user.pk]["data"])
        self.assertIn("characters", result[self.user.pk]["data"])
        self.assertIn("skills_list", result[self.user.pk]["data"])



# Standard Library
import json
from hashlib import md5

# Django
from django.apps import apps
from django.contrib.auth.models import User
from django.core.cache import cache
from django.urls import reverse
from django.utils.html import format_html

# Alliance Auth
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from memberaudit.models import CharacterSkill

# AA Memberaudit Doctrine Checker
from madc import __title__
from madc.models.skillchecker import SkillList

SKILL_CACHE_TIMEOUT_SECONDS = 60 * 60 * 72  # 72 hours
SKILL_CACHE_HEADERS_KEY = "SKILL_HEADER"
SKILL_CACHE_USER_KEY = "SKILL_LISTS_{}"

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def fittings_installed() -> bool:
    return apps.is_installed(app_name="fittings")


class SkillListHandler:
    """handler for skill lists."""

    def _get_skill_list_hash(self, skills: list[str]) -> str:
        """Generate a hash for the skill list names."""
        return md5(",".join(str(x) for x in sorted(skills)).encode()).hexdigest()

    def _get_chars_hash(self, characters: list[int]) -> str:
        return md5(",".join(str(x) for x in sorted(characters)).encode()).hexdigest()

    def _build_account_cache_key(self, characters):
        """Build a cache key based on the character IDs."""
        return SKILL_CACHE_USER_KEY.format(self._get_chars_hash(characters))

    def _generate_doctrine_html(
        self,
        skill_list_name: str,
        character_id: int,
        has_missing_skills: bool,
        skill_list_pk: int = None,
    ) -> str:
        """Generate HTML for doctrine buttons based on skill status."""
        html = f"<div class='doctrine-item btn-group' role='group' data-doctrine='{skill_list_name}'>"
        html += "<button type='button' "

        if has_missing_skills:
            url = reverse(
                viewname="madc:api:get_missing_skills",
                kwargs={"character_id": character_id, "pk": skill_list_pk},
            )
            html += f"class='btn btn-danger btn-sm' id='missing-{skill_list_name}-{character_id}' data-bs-toggle='modal' data-bs-target='#modalViewMissingContainer' data-ajax_missing='{url}'>"
        else:
            html += "class='btn btn-success btn-sm'>"

        html += f"{skill_list_name}</button>"
        html += "<button type='button' class='flex-one btn "

        if has_missing_skills:
            html += f"btn-danger btn-sm' id='copy-{skill_list_name}-{character_id}'>"
            html += "<i class='fa-solid fa-copy'></i>"
        else:
            html += "btn-success btn-sm'>"
            html += "<i class='fa-solid fa-check'></i>"
        html += "</button></div>"

        return format_html(html)

    def check_skill_lists(self, skill_lists: list[SkillList], linked_characters):
        """Check if the skill lists are up to date."""

        skills = (
            CharacterSkill.objects.filter(
                character__eve_character__character_id__in=linked_characters
            )
            .select_related(
                "eve_type", "eve_type__eve_group", "character__eve_character"
            )
            .order_by("eve_type__name")
        )

        skill_dict = {}
        skill_list_dict = {}
        skill_list_misc = {}

        for skill in skills:
            character = skill.character.eve_character.character_name
            skill_group = skill.eve_type.eve_group.name

            # Create a new entry for the character if it doesn't exist
            if character not in skill_dict:
                skill_dict[character] = {
                    "character_id": skill.character.eve_character.character_id,
                    "skills": {},
                }

            # Add the skill to the dictionary
            skill_dict[character]["skills"][skill.eve_type.name] = {
                "skill_group": skill_group,
                "sp_total": skill.skillpoints_in_skill,
                "active_level": skill.active_skill_level,
                "trained_level": skill.trained_skill_level,
            }

        # Get the skill lists and their skills
        for skill_list in skill_lists:
            skill_list_dict[skill_list.name] = skill_list.get_skills()
            skill_list_misc[skill_list.name] = {}
            skill_list_misc[skill_list.name]["pk"] = skill_list.pk
            skill_list_misc[skill_list.name]["order-weight"] = skill_list.ordering
            skill_list_misc[skill_list.name]["category"] = skill_list.category

        for character, character_data in skill_dict.items():
            character_data["doctrines"] = {}
            for skill_list_name, skills in skill_list_dict.items():
                character_data["doctrines"][skill_list_name] = {}
                character_data["doctrines"][skill_list_name]["skills"] = {}
                for skill, level in skills.items():
                    level = int(level)
                    if level > character_data["skills"].get(skill, {}).get(
                        "active_level", 0
                    ):
                        character_data["doctrines"][skill_list_name]["skills"][
                            skill
                        ] = level

                # Add metadata for the doctrine
                character_data["doctrines"][skill_list_name]["order"] = skill_list_misc[
                    skill_list_name
                ]["order-weight"]
                character_data["doctrines"][skill_list_name]["category"] = (
                    skill_list_misc[skill_list_name]["category"]
                )

                has_missing_skills = bool(
                    character_data["doctrines"][skill_list_name]["skills"]
                )
                character_data["doctrines"][skill_list_name]["html"] = (
                    self._generate_doctrine_html(
                        skill_list_name=skill_list_name,
                        character_id=character_data["character_id"],
                        has_missing_skills=has_missing_skills,
                        skill_list_pk=skill_list_misc[skill_list_name]["pk"],
                    )
                )
        return skill_dict

    def get_users_skill_list(self, users):
        """Get the skill list for multiple users."""
        linked_characters = CharacterOwnership.objects.filter(user__in=users).values(
            "user_id", "character__character_name", "character__character_id"
        )

        skill_lists = SkillList.objects.all().order_by("ordering", "name")
        skill_list_hash = self._get_skill_list_hash(skill_lists.values_list("name"))
        cached_header = cache.get(SKILL_CACHE_HEADERS_KEY, False)
        skill_lists_up_to_date = cached_header == skill_list_hash

        user_chars = {}
        for u in linked_characters:
            if u["user_id"] not in user_chars:
                user_chars[u["user_id"]] = {"chars": []}
            user_chars[u["user_id"]]["chars"].append(u["character__character_id"])

        for uid, c in user_chars.items():
            if skill_lists_up_to_date:
                cache_key = self._build_account_cache_key(c["chars"])
                cached_skills = cache.get(cache_key, False)

                if cached_skills is not False:  # check if cached at all?
                    cached_skills = json.loads(cached_skills)
                    if cached_skills.get("doctrines", False) != skill_list_hash:
                        c["data"] = self.get_user_skill_list(uid)
                    else:
                        c["data"] = cached_skills
                else:
                    c["data"] = self.get_user_skill_list(uid)
            else:
                c["data"] = self.get_user_skill_list(uid)

        return user_chars

    def get_user_skill_list(self, user_id: int, force_rebuild=True) -> dict:
        """Get the skill list for a user."""
        linked_characters = (
            User.objects.get(id=user_id)
            .character_ownerships.all()
            .values_list("character__character_id", flat=True)
        )

        skill_lists = SkillList.objects.all().order_by("ordering", "name")
        skill_list_hash = self._get_skill_list_hash(skill_lists.values_list("name"))
        cached_header = cache.get(SKILL_CACHE_HEADERS_KEY, False)
        skill_lists_up_to_date = cached_header == skill_list_hash

        account_key = self._build_account_cache_key(linked_characters)

        if skill_lists_up_to_date and not force_rebuild:
            cached_skills = cache.get(account_key, False)

            if cached_skills is not False:
                logger.debug(
                    "Using cached skill list for user %s: %s",
                    user_id,
                    account_key,
                )
                cached_skills = json.loads(cached_skills)
                if cached_skills.get("doctrines", False) == skill_list_hash:
                    return cached_skills

        output = {
            "doctrines": skill_list_hash,
            "characters": account_key,
            "skills_list": self.check_skill_lists(skill_lists, linked_characters),
        }

        output_json = json.dumps(output)

        cache.set(
            account_key,
            output_json,
            SKILL_CACHE_TIMEOUT_SECONDS,
        )
        cache.set(SKILL_CACHE_HEADERS_KEY, skill_list_hash)
        return output
